import json
import traceback
import configparser
import requests
import yaml
import time
from infoworks.core.iw_authentication import get_bearer_token
from infoworks.sdk.utils import IWUtils
from infoworks.sdk.url_builder import list_sources_url, list_domains_url, create_pipeline_url, create_data_connection, \
    configure_pipeline_url, get_pipeline_jobs_url, pipeline_version_base_url
from infoworks.sdk.cicd.upload_configurations.domains import Domain
from infoworks.sdk.cicd.upload_configurations.update_configurations import InfoworksDynamicAccessNestedDict
from infoworks.sdk.cicd.upload_configurations.local_configurations import PRE_DEFINED_MAPPINGS
from infoworks.sdk.local_configurations import POLLING_FREQUENCY_IN_SEC


def create_sql_pipeline_version(pipeline_client_obj, pipeline_id=None, domain_id=None, sql_query="",
                                pipeline_parameters=None):
    pipeline_version_id = None
    if pipeline_parameters is None:
        pipeline_parameters = []
    try:
        pipeline_id = str(pipeline_id)
        pipeline_client_obj.logger.info(
            'Pipeline {id} has been created/found under domain {domain_id}.'.format(id=pipeline_id,
                                                                                    domain_id=domain_id))
        create_pipeline_version_url = pipeline_version_base_url(pipeline_client_obj.client_config,
                                                                domain_id, pipeline_id)
        pv_response = pipeline_client_obj.call_api("POST", create_pipeline_version_url, {
            'Authorization': 'Bearer ' + pipeline_client_obj.client_config["bearer_token"],
            'Content-Type': 'application/json'}, data={
            "pipeline_id": pipeline_id,
            "type": "sql",
            "query": sql_query,
            "pipeline_parameters": pipeline_parameters
        })
        parsed_response = IWUtils.ejson_deserialize(pv_response.content)
        if parsed_response.status_code == 406:
            pv_response = pipeline_client_obj.call_api("POST", create_pipeline_version_url, {
                'Authorization': 'Bearer ' + pipeline_client_obj.client_config["bearer_token"],
                'Content-Type': 'application/json'}, data={
                "pipeline_id": pipeline_id,
                "type": "sql",
                "query": sql_query,
                "pipeline_parameters": pipeline_parameters
            })
            parsed_response = IWUtils.ejson_deserialize(pv_response.content)

        if parsed_response["result"].get("status", "") == "success":
            pipeline_version_id = pv_response["result"]["entity_id"]
            pipeline_client_obj.logger.info(f"Pipeline version {pipeline_version_id} created")
            # Make the new created version as active
            pv_active_response = IWUtils.ejson_deserialize(pipeline_client_obj.call_api("POST",
                                                                                        create_pipeline_version_url + f"/{pipeline_version_id}/set-active",
                                                                                        IWUtils.get_default_header_for_v3(
                                                                                            pipeline_client_obj.client_config[
                                                                                                'bearer_token'])).content)

            if pv_active_response["result"].get("status", "") != "success":
                pipeline_client_obj.logger.error(
                    f"Unable to set the pipeline version {pipeline_version_id} as active")
                pipeline_client_obj.logger.error(str(pv_active_response))
            else:
                pipeline_client_obj.logger.info(f"Pipeline version {pipeline_version_id} set as active")
        else:
            pipeline_client_obj.logger.error(f"Unable to create the new pipeline version")
            pipeline_client_obj.logger.error(str(pv_response))

    except Exception as e:
        pipeline_client_obj.logger.error('Response from server: ' + str(e))
        pipeline_client_obj.logger.exception('Error occurred while trying to create a new sql pipeline.')

    return pipeline_version_id


class Pipeline:
    def __init__(self, pipeline_config_path, environment_id, storage_id, interactive_id,
                 replace_words="", secrets=None):
        self.storage_id = storage_id
        self.interactive_id = interactive_id
        self.environment_id = environment_id
        self.secrets = secrets
        with open(pipeline_config_path, 'r') as file:
            json_string = file.read()
            if replace_words != "":
                for key, value in [item.split("->") for item in replace_words.split(";")]:
                    json_string = json_string.replace(key, value)
        self.configuration_obj = IWUtils.ejson_deserialize(json_string)

    def update_mappings_for_configurations(self, mappings):
        config = configparser.ConfigParser()
        config.read_dict(mappings)
        d = InfoworksDynamicAccessNestedDict(self.configuration_obj)
        for section in config.sections():
            if section in PRE_DEFINED_MAPPINGS:
                continue
            try:
                final = d.setval(section.split("$"), dict(config.items(section)))
            except KeyError as e:
                pass
        self.configuration_obj = d.data
        # handle domain name mappings
        iw_mappings = self.configuration_obj.get("configuration", {}).get("iw_mappings", [])
        try:
            if "domain_name_mappings" in config.sections():
                domain_mappings = dict(config.items("domain_name_mappings"))
                if domain_mappings != {}:
                    for mapping in iw_mappings:
                        domain_name = mapping.get("recommendation", {}).get("domain_name", "")
                        if domain_name != "" and domain_mappings != {}:
                            mapping["recommendation"]["domain_name"] = domain_mappings.get(domain_name.lower(),
                                                                                           domain_name)
                    self.configuration_obj["configuration"]["iw_mappings"] = iw_mappings
        except Exception as e:
            print("Failed while doing the domain mappings")
            print(str(e))
            print(traceback.format_exc())
        # handle any other generic name mappings like iw_mappings$recommendation$source_name
        try:
            generic_mappings = [i for i in config.sections() if i.lower().startswith("iw_mappings$")]
            for mapping in generic_mappings:
                lineage_list = mapping.split("$")
                lineage_list.remove("iw_mappings")
                if "recommendation" in lineage_list:
                    lineage_list.remove("recommendation")
                artifact_type = lineage_list[0]
                artifact_mappings = dict(config.items(f"iw_mappings$recommendation${artifact_type}"))
                if artifact_mappings != {}:
                    for mapping in iw_mappings:
                        artifact_name = mapping.get("recommendation", {}).get(artifact_type, "")
                        if artifact_name != "" and artifact_mappings != {}:
                            mapping["recommendation"][artifact_type] = artifact_mappings.get(
                                artifact_name.lower(),
                                artifact_name)
                    self.configuration_obj["configuration"]["iw_mappings"] = iw_mappings
        except Exception as e:
            print("Failed while doing the generic mappings")
            print(str(e))
            print(traceback.format_exc())

    def poll_pipeline_job(self, pipeline_client_obj, domain_id, pipeline_id):
        list_pipeline_job_url = get_pipeline_jobs_url(config=pipeline_client_obj.client_config, domain_id=domain_id,
                                                      pipeline_id=pipeline_id)
        list_pipeline_job_url = list_pipeline_job_url + "?sort_by=\"createdAt\"&limit=1&order_by=desc"
        pipeline_client_obj.logger.info(f"list pipeline url {list_pipeline_job_url}")
        response = pipeline_client_obj.call_api("GET", list_pipeline_job_url, {
            'Authorization': 'Bearer ' + pipeline_client_obj.client_config["bearer_token"],
            'Content-Type': 'application/json'}, data=None)
        parsed_response = IWUtils.ejson_deserialize(
            response.content)
        if response.status_code == 200 and "result" in parsed_response:
            result = parsed_response.get("result", [])
            if result:
                result = result[0]
                job_id = result.get("id", None)
                job_status = result.get("status", None)
                print(f"pipeline build metadata status: {job_status}")
                if job_id is not None:
                    if job_status.lower() not in ["completed", "failed", "aborted", "canceled"]:
                        time.sleep(POLLING_FREQUENCY_IN_SEC)
                        return self.poll_pipeline_job(pipeline_client_obj, domain_id, pipeline_id)
                    else:
                        return "SUCCESS" if job_status.lower() == "completed" else "FAILED"
            else:
                print("Error during polling of pipeline build metadata job")
                print(parsed_response)
                return "FAILED"
        else:
            print("Error during polling of pipeline build metadata job")
            print(parsed_response)
            return "FAILED"

    def create(self, pipeline_client_obj, domain_id, domain_name):
        pipeline_name = self.configuration_obj["configuration"]["entity"]["entity_name"]
        sources_in_pipelines = []
        for item in self.configuration_obj["configuration"]["iw_mappings"]:
            if item["entity_type"] == "table" and "source_name" in item["recommendation"]:
                sources_in_pipelines.append(item["recommendation"].get("source_name"))
        filter_condition = IWUtils.ejson_serialize({"name": {"$in": sources_in_pipelines}})
        source_list_url = list_sources_url(pipeline_client_obj.client_config)
        pipeline_client_obj.logger.info(f"Listing source url {source_list_url}")
        src_list_url = source_list_url + f"?filter={{filter_condition}}".format(filter_condition=filter_condition)
        response = pipeline_client_obj.call_api("GET", src_list_url, {
            'Authorization': 'Bearer ' + pipeline_client_obj.client_config["bearer_token"],
            'Content-Type': 'application/json'}, data=None)
        parsed_response = IWUtils.ejson_deserialize(
            response.content)
        temp_src_ids = []
        if response.status_code == 200 and "result" in parsed_response:
            result = parsed_response.get("result", {})
            for item in result:
                temp_src_ids.append(item["id"])
        sourceids_in_pipelines = list(set(temp_src_ids))
        user_email = self.configuration_obj["user_email"]
        domain_obj = Domain(self.environment_id)
        new_pipeline_id = ''
        pipeline_json_object = {
            "name": pipeline_name,
            "environment_id": self.environment_id,
            "domain_id": domain_id
        }
        # 5.2.x versions need storage id and compute id
        batch_engine = self.configuration_obj["configuration"].get("pipeline_configs", {}).get("batch_engine", "")
        storage_id = self.storage_id
        if storage_id:
            pipeline_json_object["storage_id"] = storage_id
        if batch_engine != "":
            pipeline_json_object["batch_engine"] = batch_engine

        # 5.3.x onwards CDW support
        if self.interactive_id is None:
            pipeline_json_object["run_job_on_data_plane"] = False
        else:
            pipeline_json_object["compute_id"] = self.interactive_id

        warehouse = self.configuration_obj["configuration"].get("entity", {}).get("warehouse", "")
        if warehouse:
            pipeline_json_object["snowflake_warehouse"] = warehouse
        if domain_id is None and domain_name is None:
            pipeline_client_obj.logger.error('Either domainId or domain Name is required to create pipeline.')
            print('Either domainId or domain Name is required to create pipeline.')
            traceback.print_stack()
            raise Exception("Either domainId or domain Name is required to create pipeline.")
        if domain_name is not None and domain_id is None:
            domains_url_base = list_domains_url(pipeline_client_obj.client_config)
            filter_condition = IWUtils.ejson_serialize({"name": domain_name})
            domains_url = domains_url_base + f"?filter={{filter_condition}}".format(filter_condition=filter_condition)
            response = requests.request("GET", domains_url, headers={
                'Authorization': 'Bearer ' + pipeline_client_obj.client_config["bearer_token"],
                'Content-Type': 'application/json'}, verify=False)
            if response.status_code == 406:
                headers = pipeline_client_obj.regenerate_bearer_token_if_needed(
                    {'Authorization': 'Bearer ' + pipeline_client_obj.client_config["bearer_token"],
                     'Content-Type': 'application/json'})
                response = requests.request("GET", domains_url, headers=headers, verify=False)
            final_domain_id = None
            if response is not None:
                result = response.json().get("result", [])
                if len(result) > 0:
                    final_domain_id = result[0]["id"]
                    pipeline_json_object["domain_id"] = final_domain_id
                else:
                    pipeline_client_obj.logger.error('Can not find domain with given name {} '.format(domain_name))
                    pipeline_client_obj.logger.info('Creating a domain with given name {} '.format(domain_name))
                    print(f'Can not find domain with given name {domain_name}')
                    print(f'Creating a domain with given name {domain_name}')
                    domain_id_new = domain_obj.create(pipeline_client_obj, domain_name)
                    print('New domain id' + domain_id_new)
                    pipeline_json_object["domain_id"] = domain_id_new
                    final_domain_id = domain_id_new
            pipeline_client_obj.logger.info('domainId {}'.format(final_domain_id))
        else:
            final_domain_id = domain_id
            pipeline_json_object["domain_id"] = domain_id
        pipeline_client_obj.logger.info('Adding user {} to domain {}'.format(user_email, final_domain_id))
        print(f"Adding user {user_email} to domain {final_domain_id}")
        domain_obj.add_user_to_domain(pipeline_client_obj, final_domain_id, None, user_email)
        pipeline_client_obj.logger.info(
            'Adding sources {} to domain {}'.format(sourceids_in_pipelines, final_domain_id))
        print(f'Adding sources {sourceids_in_pipelines} to domain {final_domain_id}')
        domain_obj.add_sources_to_domain(pipeline_client_obj, final_domain_id, sourceids_in_pipelines)
        url_for_creating_pipeline = create_pipeline_url(pipeline_client_obj.client_config,
                                                        pipeline_json_object["domain_id"])
        pipeline_client_obj.logger.info('url - ' + url_for_creating_pipeline)
        json_string = IWUtils.ejson_serialize(pipeline_json_object)
        pipeline_client_obj.logger.info(json_string)
        print(json_string)
        if json_string is not None:
            try:
                response = requests.post(url_for_creating_pipeline, data=json_string, headers={
                    'Authorization': 'Bearer ' + pipeline_client_obj.client_config["bearer_token"],
                    'Content-Type': 'application/json'}, verify=False)
                if response.status_code == 406:
                    headers = pipeline_client_obj.regenerate_bearer_token_if_needed(
                        {'Authorization': 'Bearer ' + pipeline_client_obj.client_config["bearer_token"],
                         'Content-Type': 'application/json'})
                    response = requests.post(url_for_creating_pipeline, data=json_string, headers=headers, verify=False)

                response = IWUtils.ejson_deserialize(response.content)
                result = response.get('result', None)
                pipeline_client_obj.logger.info("result is: " + str(result))
                if result is None:
                    pipeline_client_obj.logger.info(
                        'Cant create pipeline. {} {}'.format(response.get('message'), response.get('details')))
                    pipeline_client_obj.logger.info('Getting the existing pipelineId with given name.')
                    pipeline_base_url = create_pipeline_url(pipeline_client_obj.client_config, final_domain_id)
                    filter_condition = IWUtils.ejson_serialize({"name": pipeline_name})
                    pipeline_get_url = pipeline_base_url + f"?filter={{filter_condition}}".format(
                        filter_condition=filter_condition)
                    response = requests.request("GET", pipeline_get_url, headers={
                        'Authorization': 'Bearer ' + pipeline_client_obj.client_config["bearer_token"],
                        'Content-Type': 'application/json'}, verify=False)
                    if response.status_code == 406:
                        headers = pipeline_client_obj.regenerate_bearer_token_if_needed(
                            {'Authorization': 'Bearer ' + pipeline_client_obj.client_config["bearer_token"],
                             'Content-Type': 'application/json'})
                        response = requests.request("GET", pipeline_get_url, headers=headers, verify=False)
                    existing_pipeline_id = None
                    if response.status_code == 200 and len(response.json().get("result", [])) > 0:
                        existing_pipeline_id = response.json().get("result", [])[0]["id"]
                    if existing_pipeline_id:
                        new_pipeline_id = str(existing_pipeline_id)
                else:
                    new_pipeline_id = result.get('id')
                pipeline_client_obj.logger.info(f'Pipeline ID: {new_pipeline_id}')
                print(f'Pipeline ID: {new_pipeline_id}')
            except Exception as ex:
                pipeline_client_obj.logger.exception('Response from server: {}'.format(str(ex)))
                print(f'Response from server: {str(ex)}')

        return new_pipeline_id, pipeline_json_object["domain_id"]

    def configure(self, pipeline_client_obj, pipeline_id, domain_id, override_dataconnection_config_file=None,
                  mappings=None, read_passwords_from_secrets=False, env_tag="", secret_type=""):
        if mappings is None:
            mappings = {}
        if self.configuration_obj.get("dataconnection_configurations", None):
            pipeline_client_obj.logger.info("Checking for any data connection")
            print("Checking for any data connection")
            create_data_connection_url = create_data_connection(pipeline_client_obj.client_config, domain_id)
            for item in self.configuration_obj.get("dataconnection_configurations"):
                pipeline_client_obj.logger.info("Creating a data connection {}".format(item["name"]))
                print(f"Creating a data connection {item['name']}")
                override_keys = []
                if override_dataconnection_config_file is not None:
                    with open(override_dataconnection_config_file) as file:
                        information = yaml.load(file, Loader=yaml.FullLoader)
                    if information["dataconnection_details"].get(item["name"], None) is not None:
                        override_keys = information["dataconnection_details"].get(item["name"], {}).keys()
                        for key in override_keys:
                            item["properties"][key] = information["source_details"][item["name"]][key]
                # update the bigquery service json if any for data connection
                if item.get("sub_type").upper() == "BIGQUERY":
                    if item.get("properties").get("authentication_mechanism",
                                                  "").lower() != "system" and "server_path" not in override_keys:

                        server_path = item.get("properties").get("server_path")
                        if "gcp_details" in mappings:
                            server_path = mappings["gcp_details"].get("service_json_path")
                        if "service_json_mappings" in mappings:
                            server_path = mappings["service_json_mappings"].get(
                                server_path.split("/")[-1],
                                server_path)
                        item["properties"]["server_path"] = server_path
                    if "upload_option" in item["properties"]:
                        item["properties"]["upload_option"] = "serverLocation"
                elif item.get("sub_type").upper() in ["SNOWFLAKE", "POSTGRES"]:
                    pass
                data = IWUtils.ejson_serialize(item)
                response = requests.post(create_data_connection_url,
                                         headers={'Authorization': 'Bearer ' + pipeline_client_obj.client_config[
                                             "bearer_token"],
                                                  'Content-Type': 'application/json'}, data=data, verify=False)
                if response.status_code == 406:
                    pipeline_client_obj.client_config['bearer_token'] = get_bearer_token(
                        pipeline_client_obj.client_config["protocol"],
                        pipeline_client_obj.client_config["ip"],
                        pipeline_client_obj.client_config["port"],
                        pipeline_client_obj.client_config["refresh_token"])
                    headers = IWUtils.get_default_header_for_v3(pipeline_client_obj.client_config['bearer_token'])
                    response = requests.post(create_data_connection_url, headers=headers, data=data, verify=False)
                response = IWUtils.ejson_deserialize(response.content)
                pipeline_client_obj.logger.info(response)
                print(response)

        import_configs = {
            "run_pipeline_metadata_build": True,
            "is_pipeline_version_active": True,
            "import_data_connection": True,
            "include_optional_properties": True
        }
        url_for_importing_pipeline = configure_pipeline_url(pipeline_client_obj.client_config, domain_id, pipeline_id)
        pipeline_client_obj.logger.info(f"URL to configure pipeline: {url_for_importing_pipeline}")
        print(f"URL to configure pipeline: {url_for_importing_pipeline}")
        del self.configuration_obj["environment_configurations"]
        del self.configuration_obj["user_email"]
        json_string = IWUtils.ejson_serialize(
            {"configuration": self.configuration_obj["configuration"], "import_configs": import_configs})

        response = requests.post(url_for_importing_pipeline, data=json_string,
                                 headers={
                                     'Authorization': 'Bearer ' + pipeline_client_obj.client_config["bearer_token"],
                                     'Content-Type': 'application/json'}, verify=False)
        if response.status_code == 406:
            pipeline_client_obj.client_config['bearer_token'] = get_bearer_token(
                pipeline_client_obj.client_config["protocol"],
                pipeline_client_obj.client_config["ip"], pipeline_client_obj.client_config["port"],
                pipeline_client_obj.client_config["refresh_token"])
            headers = IWUtils.get_default_header_for_v3(pipeline_client_obj.client_config['bearer_token'])
            response = requests.post(url_for_importing_pipeline, data=json_string, headers=headers, verify=False)
        response = IWUtils.ejson_deserialize(response.content)
        if response is not None:
            pipeline_client_obj.logger.info(response.get("message", "") + " Done")
            pipeline_client_obj.logger.info(response)
            print(f'{response.get("message", "")} Done')
            print(response)
            pipeline_metadata_build_status = self.poll_pipeline_job(pipeline_client_obj, domain_id, pipeline_id)
            return pipeline_metadata_build_status
        else:
            print(response)
            return "FAILED"



import traceback

from infoworks.sdk import url_builder
from infoworks.sdk.base_client import BaseClient
from infoworks.sdk.cicd.upload_configurations.workflow import Workflow
from pathlib import Path
from infoworks.sdk.generic_response import GenericResponse
from infoworks.sdk.utils import IWUtils
from infoworks.sdk.local_configurations import Response
import os.path
import queue
import threading
import networkx as nx


class WrapperWorkflow(BaseClient):
    def __init__(self):
        super().__init__()

    def __wrapper_get_environment_details(self, environment_id=None, params=None):
        if params is None and environment_id is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_environments = url_builder.get_environment_details(
            self.client_config, environment_id) + IWUtils.get_query_params_string_from_dict(params=params)
        env_details = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_environments,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                if environment_id is not None:
                    env_details.extend(result)
                else:
                    while len(result) > 0:
                        env_details.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=env_details)
        except Exception as e:
            self.logger.error("Error in getting environment details")
            print("Error in getting environment details")

    def __wrapper_get_storage_details(self, environment_id, storage_id=None, params=None):
        if params is None and storage_id is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_storages = url_builder.get_environment_storage_details(
            self.client_config, environment_id, storage_id) + IWUtils.get_query_params_string_from_dict(params=params)
        storage_details = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_storages,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                if storage_id is not None:
                    storage_details.extend(result)
                else:
                    while len(result) > 0:
                        storage_details.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=storage_details)
        except Exception as e:
            self.logger.error("Error in getting storage details")
            print("Error in getting storage details")

    def __wrapper_get_compute_template_details(self, environment_id, compute_id=None, is_interactive=False,
                                               params=None):
        if params is None and compute_id is None:
            params = {"limit": 20, "offset": 0}
        if is_interactive:
            url_to_list_computes = url_builder.get_environment_interactive_compute_details(
                self.client_config, environment_id, compute_id) + IWUtils.get_query_params_string_from_dict(
                params=params)
        else:
            url_to_list_computes = url_builder.get_environment_compute_details(
                self.client_config, environment_id, compute_id) + IWUtils.get_query_params_string_from_dict(
                params=params)
        compute_details = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_computes,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None and "result" in response:
                result = response.get("result", [])
                if compute_id is not None:
                    compute_details.extend(result)
                else:
                    while len(result) > 0:
                        compute_details.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
            return GenericResponse.parse_result(status=Response.Status.SUCCESS, response=compute_details)
        except Exception as e:
            self.logger.error("Error in getting compute template details")
            print("Error in getting compute template details")

    def cicd_create_configure_workflow(self, configuration_file_path, domain_id=None, domain_name=None,
                                       replace_words=""):
        """
        Function to create and configure workflow using the workflow configuration JSON file.
        Pass either domain_id or domain_name.If both are not passed the name of the domain should be first part of the file name
        :param configuration_file_path: Path of the file with workflow configurations to be imported
        :param domain_id: Domain id to which the pipeline belongs to.
        :param domain_name: Domain name to which the pipeline belongs to
        :param replace_words: Pass the strings to be replaced in the configuration file. Example: DEV->PROD;dev->prod
        """
        try:
            if domain_id is None and domain_name is None:
                domain_name = Path(configuration_file_path).name.split("#")[0]
                domain_mappings = self.mappings.get("domain_name_mappings", {})
                if domain_mappings != {}:
                    if domain_name.lower() in domain_mappings.keys():
                        domain_name = domain_mappings.get(domain_name.lower(),domain_name)
            wf_obj = Workflow(configuration_file_path, replace_words)
            wf_obj.update_mappings_for_configurations(self.mappings)
            wf_id, domain_id = wf_obj.create(self, domain_id, domain_name)
            if wf_id is not None:
                status = wf_obj.configure(self, wf_id, domain_id)
        except Exception as e:
            self.logger.error(str(e))
            print(str(e))
            print(traceback.format_exc())

    def __execute(self, thread_number, q):
        while True:
            try:
                print('%s: Looking for the next task ' % thread_number)
                task = q.get()
                print(f'\nThread Number {thread_number} processed {task}')
                entity_type = task["entity_type"]
                if entity_type == "workflow":
                    replace_words = task["replace_words"] if task["replace_words"] else ""
                    self.cicd_create_configure_workflow(task["workflow_config_path"], None,
                                                        task["domain_name"],
                                                        replace_words)
                else:
                    pass
            except Exception as e:
                print(str(e))
            q.task_done()

    def cicd_create_workflowartifacts_in_bulk(self, maintain_lineage, workflow_gexf_lineage_file=None,
                           configurations_base_path=None,
                           wf_replace_words=None):
        num_fetch_threads = 10
        job_queue = queue.Queue(maxsize=100)
        for i in range(num_fetch_threads):
            worker = threading.Thread(target=self.__execute, args=(i, job_queue,))
            worker.setDaemon(True)
            worker.start()
        if maintain_lineage == 'true':
            if workflow_gexf_lineage_file:
                path = os.path.join(configurations_base_path, "modified_files/workflow.gexf")
            else:
                path = workflow_gexf_lineage_file
            graph = nx.read_gexf(path)
            for item in self._topological_sort_grouping(graph):
                for i in item:
                    if i != "root":
                        wf_name = graph._node[i]['filename'].strip()
                        domain_name = wf_name.split("#")[0]
                        domain_mappings = self.mappings.get("domain_name_mappings", {})
                        if domain_mappings != {}:
                            if domain_name.lower() in domain_mappings.keys():
                                domain_name = domain_mappings.get(domain_name.lower(), domain_name)
                        wf_args = {"entity_type": "workflow",
                                   "workflow_config_path": os.path.join("configurations/workflow", wf_name),
                                   "domain_name": domain_name}
                        job_queue.put(wf_args)
                print('*** Main thread waiting to complete all pending pipeline creation/configuration requests ***')
                job_queue.join()
                print('*** Done with All Tasks ***')
        else:
            with open(os.path.join("configurations/modified_files", "workflow.csv"), "r") as workflow_file_fp:
                for workflow_file in workflow_file_fp.readlines():
                    wf_name = workflow_file.strip()
                    domain_name = wf_name.split("#")[0]
                    domain_mappings = self.mappings.get("domain_name_mappings", {})
                    if domain_mappings != {}:
                        if domain_name.lower() in domain_mappings.keys():
                            domain_name = domain_mappings.get(domain_name.lower(), domain_name)
                    wf_args = {"entity_type": "workflow",
                               "workflow_config_path": os.path.join("configurations/workflow", wf_name),
                               "domain_name": domain_name, "replace_words": wf_replace_words, }
                    job_queue.put(wf_args)

            print('*** Main thread waiting to complete all workflow configuration requests ***')
            job_queue.join()
            print('*** Done with Workflow Configurations  ***')


import time

from infoworks.error import WorkflowError
from infoworks.sdk import url_builder
from infoworks.sdk.base_client import BaseClient
from infoworks.sdk.local_configurations import Response, ErrorCode
from infoworks.sdk.utils import IWUtils
from infoworks.sdk.workflow_response import WorkflowResponse


class WorkflowClient(BaseClient):
    def __init__(self):
        super(WorkflowClient, self).__init__()

    def get_list_of_workflows(self, domain_id=None, params=None):
        """
        Gets List of Infoworks Data workflow details for given domain id
        :param domain_id: Domain id to which the workflows belongs to, if None all workflows in all domains will be fetched
        :type domain_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response dict
        """
        response = None
        initial_msg = ""
        try:
            if domain_id is None:
                if params is None:
                    params = {"limit": 20, "offset": 0}
                url_to_list_workflows = url_builder.get_all_workflows_url(
                    self.client_config) + IWUtils.get_query_params_string_from_dict(params=params)
                workflows_list = []
                response = IWUtils.ejson_deserialize(
                    self.call_api("GET", url_to_list_workflows,
                                  IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
                if response is not None:
                    initial_msg = response.get("message", "")
                    result = response.get("result", {}).get("items", [])
                    while len(result) > 0:
                        workflows_list.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", {}).get("items", [])
                response["result"] = workflows_list
                response["message"] = initial_msg
                return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, response=response)
            else:
                workflows_list = []
                response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.create_workflow_url(
                    self.client_config, domain_id), IWUtils.get_default_header_for_v3(
                    self.client_config['bearer_token'])).content)

                if response is not None:
                    initial_msg = response.get("message", "")
                    result = response.get("result", [])
                    while len(result) > 0:
                        workflows_list.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
                response["result"] = workflows_list
                response["message"] = initial_msg
                return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to get workflow details.')
            raise WorkflowError('Error occurred while trying to get workflow details.')

    def create_workflow(self, domain_id=None, workflow_config=None):
        """
        Create a new Workflow
        :param domain_id: Domain id of the workflow
        :type domain_id:String
        :param workflow_config: a JSON object containing workflow configurations
        :type workflow_config: JSON Object
        ```
        workflow_config_example = {
        "name": "workflow_name"
        }
        ```
        :return: response dict
        """
        if None in {domain_id} and workflow_config is None:
            self.logger.error("domain id or workflow_config cannot be None")
            raise Exception("domain_id id or workflow_config cannot be None")
        response = None
        try:
            response = IWUtils.ejson_deserialize(self.call_api("POST", url_builder.create_workflow_url(
                self.client_config, domain_id), IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                                               workflow_config).content)

            result = response.get('result', {})
            workflow_id = result.get('id', None)

            if workflow_id is None:
                self.logger.error('Workflow failed to create.')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc='Workflow failed to create.', response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Workflow {id} has been created under domain {domain_id}.'.format(id=workflow_id, domain_id=domain_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to create a new workflow.')
            raise WorkflowError('Error occurred while trying to create a new workflow.')

    def get_workflow_details(self, workflow_id=None, domain_id=None):
        """
        Gets Infoworks Data workflow details for given workflow id
        :param workflow_id: id of the workflow whose details are to be fetched
        :type workflow_id: String
        :param domain_id: Domain id to which the workflow belongs to
        :type domain_id: String
        :return: response dict
        """
        if None in {domain_id, workflow_id}:
            self.logger.error("domain id or workflow_id cannot be None")
            raise Exception("domain_id id or workflow_id cannot be None")
        response = None
        try:
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.create_workflow_url(
                self.client_config, domain_id) + f"/{workflow_id}", IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)

            result = response.get('result', {})

            if result.get('id', None) is None:
                self.logger.error('Failed to find the workflow details')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc='Failed to find the workflow details',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully got the workflow {id} details.'.format(id=workflow_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to get workflow details.')
            raise WorkflowError('Error occurred while trying to get workflow details.')

    def delete_workflow(self, workflow_id=None, domain_id=None):
        """
        Deletes Infoworks Data workflow  for given workflow id
        :param workflow_id: entity id of the workflow to be deleted
        :type workflow_id: String
        :param domain_id: Domain id to which the workflow belongs to
        :type domain_id: String
        :return: response dict
        """
        if None in {domain_id, workflow_id}:
            self.logger.error("domain id or workflow_id cannot be None")
            raise Exception("domain_id id or workflow_id cannot be None")
        response = None
        try:
            if workflow_id is None:
                self.logger.error('Please pass the mandatory workflow id as parameter.')
                raise WorkflowError('Please pass the mandatory workflow_id as parameter.')
            if domain_id is None:
                self.logger.error('Please pass the mandatory domain_id as parameter.')
                raise WorkflowError('Please pass the mandatory domain_id as parameter.')
            response = IWUtils.ejson_deserialize(self.call_api("DELETE", url_builder.create_workflow_url(
                self.client_config, domain_id) + f"/{workflow_id}", IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)

            result = response.get('result', {})

            if result.get('id', None) is None:
                self.logger.error(f'Failed to delete the workflow {workflow_id}')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc=f'Failed to delete the workflow {workflow_id}',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully deleted the workflow {id}.'.format(id=workflow_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to delete workflow.')
            raise WorkflowError('Error occurred while trying to delete workflow.')

    def update_workflow(self, workflow_id=None, domain_id=None, workflow_config=None):
        """
        Updates Infoworks Data workflow details for given workflow id
        :param workflow_id: entity id of the workflow to be updated
        :type workflow_id: String
        :param domain_id: Domain id to which the workflow belongs to
        :type domain_id: String
        :param workflow_config: a JSON object containing workflow configurations
        :type workflow_config: JSON Object
        ```
        workflow_config_example = {
         "name": "{string}",
         "description": "{string}",
         "child_workflow_ids": [
          "{array[string]...}"
         ],
         "domainId": "{string}",
         "workflow_graph": {
          "tasks": [
           {
            "task_id": "e77850ad5127a2d7dab870ff",
            "task_type": "ingest_table_group",
            "location": "-237 52",
            "title": "ingest",
            "task_properties": {},
            "run_properties": {
             "enable_exponential_backoff_for_retries": true,
             "num_retries": 5,
             "retry_delay": 200,
             "max_retry_delay": 1000,
             "trigger_rule": "{string}"
            },
            "workflow_variables": "{}"
           }
          ],
          "edges": [
           {
            "category": "{string}",
            "from_task": "{string}",
            "to_task": "{string}"
           }
          ]
         }
        }
        ```
        :return: response dict
        """
        if None in {domain_id, workflow_id} and workflow_config is None:
            self.logger.error("domain id or workflow_id or workflow_config cannot be None")
            raise Exception("domain_id or workflow_id or workflow_config cannot be None")
        response = None
        try:
            if "childWorkflowIds" not in workflow_config:
                workflow_config["child_workflow_ids"] = []
            response = IWUtils.ejson_deserialize(self.call_api("PATCH", url_builder.create_workflow_url(
                self.client_config, domain_id) + f"/{workflow_id}", IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token']), workflow_config).content)

            result = response.get('result', {})

            if result.get('id', None) is None:
                self.logger.error(f'Failed to update the workflow {workflow_id}')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc=f'Failed to update the workflow {workflow_id}',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully updated the workflow {id}.'.format(id=workflow_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to update workflow.')
            raise WorkflowError('Error occurred while trying to update workflow.')

    def trigger_workflow(self, workflow_id=None, domain_id=None, trigger_wf_body=None):
        """
        Triggers Infoworks Data workflow for given workflow id
        :param workflow_id: entity id of the workflow to be triggered
        :type workflow_id: String
        :param domain_id: Domain id to which the workflow belongs to
        :type domain_id: String
        :param trigger_wf_body: Pass the workflow parameters if any
        ```
        trigger_wf_body = {
            "workflow_parameters": [
                {
                    "key": "name",
                    "value": "WF_API_TRIGGER"
                }
            ]
        }
        ```
        :return: response dict
        """
        if None in {domain_id, workflow_id}:
            self.logger.error("domain id or workflow_id cannot be None")
            raise Exception("domain_id or workflow_id cannot be None")
        response = None
        try:
            if trigger_wf_body:
                response = IWUtils.ejson_deserialize(self.call_api("POST", url_builder.trigger_workflow_url(
                    self.client_config, domain_id, workflow_id), IWUtils.get_default_header_for_v3(
                    self.client_config['bearer_token']), data=trigger_wf_body).content)
            else:
                response = IWUtils.ejson_deserialize(self.call_api("POST", url_builder.trigger_workflow_url(
                    self.client_config, domain_id, workflow_id), IWUtils.get_default_header_for_v3(
                    self.client_config['bearer_token'])).content)
            result = response.get('result', {})
            run_id = result.get('id', None)
            if result.get('id', None) is None:
                self.logger.error(f'Failed to trigger the workflow {workflow_id}')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc=f'Failed to trigger the workflow {workflow_id}',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully triggered the workflow {id} with run id {run_id}.'.format(id=workflow_id, run_id=run_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to trigger workflow.')
            raise WorkflowError('Error occurred while trying to trigger workflow.')

    def restart_or_cancel_multiple_workflows(self, action_type="restart", workflow_list_body=None):
        """
        Restart/Cancel Infoworks Data workflow for given workflow id
        :param action_type: restart/cancel
        :param workflow_list_body: JSON object containing array of ids(workflow_id,run_id) to restart/cancel
        :type workflow_list_body: JSON Dict
        ```
        example: {
                  "ids": [
                    {
                      "workflow_id": "e77850ad5127a2d7dab870ff",
                      "run_id": "e77850ad5127a2d7dab870ff"
                    }
                  ]
                }
        ```
        :return: response dict
        """
        if workflow_list_body is None:
            self.logger.error("workflow_list_body cannot be None")
            raise Exception("workflow_list_body cannot be None")
        if action_type.lower() not in ["restart", "cancel"]:
            self.logger.error("action_type should be restart or cancel")
            raise Exception("action_type should be restart or cancel")
        response = None
        try:
            if action_type.lower() == "restart":
                request_url = url_builder.restart_multiple_workflows_url(self.client_config)
            else:
                request_url = url_builder.cancel_multiple_workflows_url(self.client_config)
            response = IWUtils.ejson_deserialize(self.call_api("POST", request_url, IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token']), data=workflow_list_body).content)

            result = response.get('result', None)
            if result is None:
                self.logger.error(f'Failed to {action_type} the workflows')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc=f'Failed to {action_type} the workflows',
                                                     response=response)

            self.logger.info(
                f'Successfully {action_type}d the workflows.')
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception(f'Error occurred while {action_type}ing the workflows.' + str(e))
            raise WorkflowError(f'Error occurred while {action_type}ing the workflows.' + str(e))

    def cancel_multiple_workflow(self, workflow_list_body=None):
        """
        Cancels Infoworks Data workflow for given workflow id
        :param workflow_list_body: JSON object containing array of ids(workflow_id,run_id) to restart
        :type workflow_list_body: JSON dict
        ```
        example: {
                  "ids": [
                    {
                      "workflow_id": "e77850ad5127a2d7dab870ff",
                      "run_id": "e77850ad5127a2d7dab870ff"
                    }
                  ]
                }
        ```
        :return: response dict
        """
        response = None
        if workflow_list_body is None:
            self.logger.error("workflow_list_body cannot be None")
            raise Exception("workflow_list_body cannot be None")
        try:
            response = IWUtils.ejson_deserialize(self.call_api("POST", url_builder.cancel_multiple_workflows_url(
                self.client_config), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token']), data=workflow_list_body).content)

            result_msg = response.get("message", None)
            if result_msg is None and result_msg == "Submitted Cancel Job for Workflow Run":
                self.logger.error(f'Failed to cancel the workflows')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc=f'Failed to cancel the workflows',
                                                     response=response)

            self.logger.info(
                'Successfully Submitted Cancel Job for Workflow Run.')
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while cancelled the workflows.' + str(e))
            raise WorkflowError('Error occurred while cancelled the workflows.' + str(e))

    def get_status_of_workflow(self, workflow_run_id=None, workflow_id=None):
        """
        Fetches status of Infoworks Data workflow for given workflow id and run id
        :param workflow_run_id: run id of the workflow running
        :type workflow_run_id: String
        :param workflow_id: entity id of the workflow running
        :type workflow_id: String
        :return: response dict
        """
        if None in {workflow_id, workflow_run_id}:
            self.logger.error("workflow_id or workflow_run_id cannot be None")
            raise Exception("workflow_id or workflow_run_id cannot be None")
        response = None
        try:
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.get_workflow_status_url(
                self.client_config, workflow_id, workflow_run_id), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)

            result = response.get('result', None)
            if result is None:
                self.logger.error('Failed to get the status of workflow')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc='Failed to get the configuration json of workflow',
                                                     response=response)
            run_id = result.get('id', None)
            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully got status of the workflow {id} with run id {run_id}.'.format(id=workflow_id,
                                                                                            run_id=run_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=result)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to fetch status of workflow.')
            raise WorkflowError('Error occurred while trying to fetch status of workflow.')

    def poll_workflow_run_till_completion(self, workflow_run_id=None, workflow_id=None, poll_interval=30):
        """
        Polls Infoworks Data workflow for given workflow id and run id
        :param workflow_run_id: run id of the workflow running
        :type workflow_run_id: String
        :param workflow_id: entity id of the workflow running
        :type workflow_id: String
        :param poll_interval:interval in seconds between poll(default 30)
        :type poll_interval : Int
        :return: response dict
        """
        response = None
        if None in {workflow_id, workflow_run_id}:
            self.logger.error("workflow_id or workflow_run_id cannot be None")
            raise Exception("workflow_id or workflow_run_id cannot be None")
        try:
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.get_workflow_status_url(
                self.client_config, workflow_id, workflow_run_id), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)
            result = response.get('result', {})
            if result:
                workflow_status = result['workflow_status']["state"]
                while workflow_status.lower() not in ['success', 'completed', 'failed', 'aborted', 'canceled']:
                    response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.get_workflow_status_url(
                        self.client_config, workflow_id, workflow_run_id), IWUtils.get_default_header_for_v3(
                        self.client_config['bearer_token'])).content)
                    result = response.get('result', {})
                    workflow_status = result['workflow_status']["state"]
                    if workflow_status.lower() in ['success', 'completed', 'failed', 'aborted', 'canceled']:
                        break
                    print(f"workflow_status : {workflow_status}.Sleeping for {poll_interval} seconds")
                    time.sleep(poll_interval)

                run_id = result.get('id', None)
                if result.get('id', None) is None:
                    self.logger.error(f'Failed to poll status of the workflow {workflow_id}')
                    return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                         error_code=ErrorCode.USER_ERROR,
                                                         error_desc=f'Failed to poll status of the workflow {workflow_id}',
                                                         response=response)

                workflow_id = str(workflow_id)
                self.logger.info(
                    'Successfully polled status of the workflow {id} with run id {run_id}.'.format(id=workflow_id,
                                                                                                   run_id=run_id))
                return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                     response=response)
            else:
                print(response)
                raise Exception("Returned API Result is None")

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to poll status of workflow.')
            raise WorkflowError('Error occurred while trying to poll status of workflow.')

    def get_workflow_configuration_json_export(self, workflow_id=None, domain_id=None):
        """
        Get exported config for workflow with workflow_id in domain with domain_id
        :param workflow_id: id of the workflow whose details is to be fetched
        :type workflow_id: String
        :param domain_id: Domain id to which the workflow belongs to
        :type domain_id: String
        :return: response dict
        """
        response = None
        if None in {workflow_id, domain_id}:
            self.logger.error("workflow_id or domain_id cannot be None")
            raise Exception("workflow_id or domain_id cannot be None")
        try:
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.configure_workflow_url(
                self.client_config, domain_id, workflow_id), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)

            result = response.get('result', None)

            if result is None:
                self.logger.error('Failed to get the configuration json of workflow')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc='Failed to get the configuration json of workflow',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully got the workflow {id} configuration json.'.format(id=workflow_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to get workflow configuration json.')
            raise WorkflowError('Error occurred while trying to get workflow configuration json.')

    def update_workflow_configuration_json_import(self, workflow_id=None, domain_id=None, workflow_config=None):
        """
        Import config for workflow with workflow_id in domain with domain_id
        :param workflow_id: id of the workflow whose details are to be fetched
        :type workflow_id: String
        :param domain_id: Domain id to which the workflow belongs to
        :type domain_id: String
        :param workflow_config: configuration json of the workflow
        :type workflow_config: JSON Object
        :return: response dict
        """
        if None in {workflow_id, domain_id} and workflow_config is None:
            self.logger.error("workflow_id or domain_id or workflow_config cannot be None")
            raise Exception("workflow_id or domain_id or workflow_config cannot be None")
        response = None
        try:
            response = IWUtils.ejson_deserialize(self.call_api("PUT", url_builder.configure_workflow_url(
                self.client_config, domain_id, workflow_id), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token']), workflow_config).content)

            result = response.get('result', None)

            if result is None:
                self.logger.error('Failed to update the configuration json of workflow')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc='Failed to update the configuration json of workflow',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully updated the workflow {id} configuration json.'.format(id=workflow_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to update workflow configuration json.')
            raise WorkflowError('Error occurred while trying to update workflow configuration json.')

    def get_workflow_id(self, workflow_name=None, domain_id=None, domain_name=None):
        """
        Function to get workflow id
        :param workflow_name: Name of the workflow
        :param domain_id: Entity identifier of the domain
        :param domain_name: Entity name of the domain
        :return: Workflow ID
        ```
        example response : {"id": "884236e85b9b1a69b2907e4c"}
        ```
        """
        if None in {workflow_name}:
            self.logger.error("workflow_name cannot be None")
            raise Exception("workflow_name cannot be None")
        if domain_id is None and domain_name is None:
            raise Exception(f"Either domain name or domain id has to be passed")
        if domain_name is not None and domain_id is None:
            # Find the domain_id
            params = {"filter": {"name": domain_name}}
            url_to_list_domains = url_builder.create_domain_url(
                self.client_config) + IWUtils.get_query_params_string_from_dict(params=params)
            try:
                response = IWUtils.ejson_deserialize(
                    self.call_api("GET", url_to_list_domains,
                                  IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
                if response is not None:
                    result = response.get("result", [])
                    domain_id = result[0]["id"]
            except Exception as e:
                self.logger.error("Error in listing domains")
                raise WorkflowError("Error in listing domains" + str(e))
        try:
            params = {"filter": {"name": workflow_name}}
            url_to_get_wf_info = url_builder.create_workflow_url(self.client_config,
                                                                 domain_id) + IWUtils.get_query_params_string_from_dict(
                params=params)
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_wf_info,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", None)
                if result is not None:
                    return WorkflowResponse.parse_result(status=Response.Status.SUCCESS,
                                                         response={"id": result[0].get('id')})
        except:
            raise WorkflowError("Unable to get workflow id from name")

    def get_workflow_name(self, domain_id=None, workflow_id=None):
        """
        Function to get workflow name from ID
        :param workflow_id: Id of the workflow
        :type workflow_id: String
        :param domain_id: Entity identifier of the domain
        :type domain_id: String
        :return: Workflow name
        ```
        example response : {"name": "sample_workflow"}
        ```
        """
        if None in {workflow_id, domain_id}:
            self.logger.error("workflow_id or domain_id cannot be None")
            raise Exception("workflow_id or domain_id cannot be None")
        try:
            url_to_get_wf_info = url_builder.create_workflow_url(self.client_config, domain_id) + f"/{workflow_id}"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_wf_info,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", None)
                if result is not None:
                    return WorkflowResponse.parse_result(status=Response.Status.SUCCESS,
                                                         response={"name": result.get("name")})
        except:
            raise WorkflowError("Unable to get workflow name")

    def get_list_of_workflow_runs(self, domain_id=None, workflow_id=None, params=None, api_body_for_filter={}):
        """
        Gets List of Infoworks Data workflow runs details for given domain id and workflow id
        :param domain_id: Domain id to which the workflows belongs to, if None all workflows in all domains will be fetched
        :type domain_id: String
        :param workflow_id: Workflow id,if None all workflow runs in all domains will be fetched
        :type workflow_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :param api_body_for_filter: dict of API body
        :type api_body_for_filter: JSON dict
        ```
        example:
        {'limit': 10000, 'date_range': {'type': 'last', 'unit': 'day', 'value': 1}, 'offset': 0}
        ```
        :return: response List
        """
        response = None
        initial_msg = ""
        try:
            if domain_id is None:
                url_to_list_workflow_runs = url_builder.get_all_workflows_runs_url(
                    self.client_config) + IWUtils.get_query_params_string_from_dict(params=params)
                if not api_body_for_filter.get("limit", None):
                    api_body_for_filter["limit"] = 10000

                if not api_body_for_filter.get("date_range", None):
                    api_body_for_filter["date_range"] = {"type": "last", "unit": "day", "value": 1}
                api_body_for_filter["offset"] = 0
                workflow_runs_list = []
                response = IWUtils.ejson_deserialize(
                    self.call_api("POST", url_to_list_workflow_runs,
                                  IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                  data=api_body_for_filter).content)
                if response is not None:
                    initial_msg = response.get("message", "")
                    result = response.get("result", [])
                    if len(result) == 0:
                        self.logger.error('Failed to update the configuration json of workflow')
                        return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                             error_code=ErrorCode.USER_ERROR,
                                                             error_desc='Failed to get the workflow run id jobs',
                                                             response=response)
                    while len(result) > 0:
                        workflow_runs_list.extend(result)
                        api_body_for_filter["offset"] = api_body_for_filter.get("limit")
                        response = IWUtils.ejson_deserialize(
                            self.call_api("POST", url_to_list_workflow_runs, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token']), data=api_body_for_filter).content)
                        result = response.get("result", [])
                response["result"] = workflow_runs_list
                response["message"] = initial_msg
                return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, response=response)
            else:
                workflow_runs_list = []
                response = IWUtils.ejson_deserialize(
                    self.call_api("GET", url_builder.get_all_workflows_runs_url_with_domain_id(
                        self.client_config, domain_id, workflow_id), IWUtils.get_default_header_for_v3(
                        self.client_config['bearer_token'])).content)

                result = response.get('result', None)
                if result is None:
                    self.logger.error('Failed to get the list of workflow runs')
                    return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                         error_code=ErrorCode.USER_ERROR,
                                                         error_desc='Failed to get the workflow run id jobs',
                                                         response=response)
                if response is not None:
                    result = response.get("result", [])
                    initial_msg = response.get("message", "")
                    while len(result) > 0:
                        workflow_runs_list.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
                    response["result"] = workflow_runs_list
                    response["message"] = initial_msg
                return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to get workflow details.')
            raise WorkflowError('Error occurred while trying to get workflow details.')

    def get_list_of_workflow_runs_jobs(self, run_id=None, params=None):
        """
         Gets List of Infoworks Data workflow runs jobs details
        :param run_id: Run id to which the workflows belongs to, if None all workflows
        :type run_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response dict
        """
        if None in {run_id}:
            self.logger.error("run_id cannot be None")
            raise Exception("run_id cannot be None")
        response = None
        try:
            workflow_run_jobs_list = []
            get_all_workflow_run_jobs_url = url_builder.get_all_workflow_run_jobs_url(self.client_config, run_id)
            if params is None:
                get_all_workflow_run_jobs_url = get_all_workflow_run_jobs_url + "?fetch_summary=true&recursive_job_search=true"
            else:
                get_all_workflow_run_jobs_url = get_all_workflow_run_jobs_url + \
                                                IWUtils.get_query_params_string_from_dict(params=params) + \
                                                "&fetch_summary=true&recursive_job_search=true"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", get_all_workflow_run_jobs_url, IWUtils.get_default_header_for_v3(
                    self.client_config['bearer_token'])).content)

            result = response.get('result', None)
            initial_msg = ""
            if result is None:
                self.logger.error('Failed to get list of workflow runs jobs')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc='Failed to get the workflow run id jobs',
                                                     response=response)

            if response is not None:
                result = response.get("result", [])
                initial_msg = response.get("message", "")
                while len(result) > 0:
                    workflow_run_jobs_list.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", [])
            response["result"] = workflow_run_jobs_list
            response["message"] = initial_msg
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to get jobs under workflow run.' + str(e))
            raise WorkflowError('Error occurred while trying to get jobs under workflow run.' + str(e))

    def get_list_of_domain_workflow_schedules(self, domain_id, params=None):
        """
        Gets List of Schedules of all Workflows belonging to the Domain
        :param domain_id: Domain Id to list the schedules configured
        :type domain_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type params: JSON Dict
        :return: Response Dict
        """
        if domain_id is None:
            self.logger.error("domain_id cannot be None")
            raise Exception("domain_id cannot be None")

        if params is None:
            params = {"limit": 20, "offset": 0}

        domain_schedules = []
        response = None
        initial_msg = ''
        try:
            domain_schedules_url = url_builder.get_domain_workflow_schedules_url(self.client_config, domain_id) + \
                                   IWUtils.get_query_params_string_from_dict(params=params)

            response = IWUtils.ejson_deserialize(self.call_api("GET", domain_schedules_url,
                                                               IWUtils.get_default_header_for_v3(
                                                                   self.client_config['bearer_token'])).content)
            if response:
                initial_msg = response.get("message", '')
                result = response.get("result", [])

                while len(result) > 0:
                    domain_schedules.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", [])
            response["result"] = domain_schedules
            response["message"] = initial_msg
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to get domain schedules .')
            raise WorkflowError('Error occurred while trying to get domain schedules.')

    def get_workflow_schedule(self, domain_id, workflow_id):
        """
        Gets Schedules of particular Workflow belonging to the Domain
        :param domain_id: Domain ID of which Workflow belongs
        :type domain_id: String
        :param workflow_id: Workflow ID to fetch schedule for.
        :type workflow_id: String
        :return: Response Dict
        """
        if domain_id is None or workflow_id is None:
            self.logger.error("domain_id or workflow_id cannot be None")
            raise Exception("domain_id or workflow_id cannot be None")

        response = None

        try:
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.get_workflow_schedule_url(
                self.client_config, domain_id, workflow_id), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)

            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to get workflow schedule.')
            raise WorkflowError('Error occurred while trying to get workflow schedule.')

    def enable_workflow_schedule(self, domain_id, workflow_id, schedule_config):
        """
        Enables Schedule of a particular Workflow belonging to the Domain
        :param domain_id: Domain ID of which Workflow belongs
        :type domain_id: String
        :param workflow_id: Workflow ID to fetch schedule for.
        :type workflow_id: String
        :param schedule_config: Schedule Configuration JSON of the Workflow
        :type schedule_config: JSON
        ```
        schedule_config_example = {
              "start_date": "02/22/2020",
              "end_date": "02/24/2020",
              "start_hour": 12,
              "start_min": 25,
              "end_hour": 17,
              "end_min": 30,
              "repeat_interval_measure": 2,
              "repeat_interval_unit": "{string}",
              "ends": true,
              "is_custom_job": true,
              "custom_job_details": {
                "starts_daily_at": "14:00",
                "ends_daily_at": "15:00",
                "repeat_interval_unit": "{string}",
                "repeat_interval_measure": 2
              },
              "repeat_on_last_day": "{boolean}",
              "specified_days": 1
        }
        ```
        :return: Response Dict
        """
        if domain_id is None or workflow_id is None or schedule_config is None:
            self.logger.error("domain_id or workflow_id or schedule_config cannot be None")
            raise Exception("domain_id or workflow_id or schedule_config cannot be None")
        response = None
        try:
            response = IWUtils.ejson_deserialize(self.call_api("PUT", url_builder.get_enable_workflow_schedule_url(
                self.client_config, domain_id, workflow_id), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token']), schedule_config).content)

            result = response.get('result', None)

            if result is None:
                self.logger.error('Failed to enable schedule of the workflow')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc='Failed to enable schedule of the workflow',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully enabled schedule for the workflow {id}.'.format(id=workflow_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to enable workflow schedule.')
            raise WorkflowError('Error occurred while trying to enable workflow schedule.')

    def disable_workflow_schedule(self, domain_id, workflow_id):
        """
        Disables Schedule of a particular Workflow belonging to the Domain
        :param domain_id: Domain ID of which Workflow belongs
        :type domain_id: String
        :param workflow_id: Workflow ID to fetch schedule for.
        :type workflow_id: String
        :return: Response Dict
        """
        if domain_id is None or workflow_id is None:
            self.logger.error("domain_id or workflow_id cannot be None")
            raise Exception("domain_id or workflow_id cannot be None")
        response = None
        try:
            response = IWUtils.ejson_deserialize(self.call_api("PUT", url_builder.get_disable_workflow_schedule_url(
                self.client_config, domain_id, workflow_id), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)

            result = response.get('result', None)

            if result is None:
                self.logger.error('Failed to disable schedule of the workflow')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc='Failed to disable schedule of the workflow',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully disabled schedule for the workflow {id}.'.format(id=workflow_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to disable workflow schedule.')
            raise WorkflowError('Error occurred while trying to disable workflow schedule.')

    def update_workflow_schedule_user(self, domain_id, workflow_id):
        """
        Changes Workflow Schedule User of a particular Workflow belonging to the Domain
        :param domain_id: Domain ID of which Workflow belongs
        :type domain_id: String
        :param workflow_id: Workflow ID to fetch schedule for.
        :type workflow_id: String
        :return: Response Dict
        """
        if domain_id is None or workflow_id is None:
            self.logger.error("domain_id or workflow_id cannot be None")
            raise Exception("domain_id or workflow_id cannot be None")
        response = None
        try:
            response = IWUtils.ejson_deserialize(self.call_api("PUT", url_builder.update_workflow_schedule_user_url(
                self.client_config, domain_id, workflow_id), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)

            result = response.get('result', None)

            if result is None:
                self.logger.error('Failed to update workflow schedule user')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc='Failed to update workflow schedule user',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully updated workflow {id} schedule user.'.format(id=workflow_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to update workflow schedule user.')
            raise WorkflowError('Error occurred while trying to update workflow schedule user.')



    def pause_or_resume_multiple_workflows(self, action_type="pause", workflow_list_body=None):
        """
        Pause/resume Infoworks Data workflow for given workflow ids
        :param action_type: pause/resume
        :param workflow_list_body: JSON object containing array of workflow ids
        :type workflow_list_body: JSON Dict
        workflow_list_body_example: {
        "workflow_ids": ["c265e25b886a1b5e09896885"]
        }
        :return: response dict
        """
        if workflow_list_body is None:
            self.logger.error("workflow_list_body cannot be None")
            raise Exception("workflow_list_body cannot be None")
        if action_type.lower() not in ["pause", "resume"]:
            self.logger.error(f"action_type cannot be {action_type}. Supported resume/pause")
            raise Exception(f"action_type cannot be {action_type}. Supported resume/pause")
        response = None
        try:
            if action_type.lower() == "pause":
                request_url = url_builder.pause_multiple_workflows_url(self.client_config)
            else:
                request_url = url_builder.resume_multiple_workflows_url(self.client_config)
            response = IWUtils.ejson_deserialize(self.call_api("POST", request_url, IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token']), data=workflow_list_body).content)
            result = response.get('result', None)
            if result is None:
                self.logger.error(f'Failed to {action_type} the workflows')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc=f'Failed to {action_type} the workflows',
                                                     response=response)

            self.logger.info(
                f'Successfully {action_type}d the workflows.')
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception(f'Error occurred while {action_type}d the workflows.' + str(e))
            raise WorkflowError(f'Error occurred while {action_type}d the workflows.' + str(e))

    def get_global_list_of_workflow_runs(self, params=None):
        """
        Gets List of all the Infoworks Data workflows. Need admin access
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response dict
        """
        response = None
        initial_msg = ""
        try:
            if params is None:
                params = {"limit": 20, "offset": 0}
            url_to_list_workflow_runs = url_builder.get_workflow_runs_url(
                self.client_config) + IWUtils.get_query_params_string_from_dict(params=params)
            workflow_run_list = []
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_workflow_runs,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                initial_msg = response.get("message", "")
                result = response.get("result", {}).get("items", [])
                while len(result) > 0:
                    workflow_run_list.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", {}).get("items", [])
            response["result"] = workflow_run_list
            response["message"] = initial_msg
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to get workflow run details.')
            raise WorkflowError('Error occurred while trying to get workflow run details.')

    def get_list_of_workflow_tasks(self, params=None):
        """
        Gets List of all the Infoworks Data workflow tasks
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response dict
        """
        response = None
        initial_msg = ""
        try:
            if params is None:
                params = {"limit": 20, "offset": 0}
            url_to_list_workflow_run_tasks = url_builder.get_workflow_tasks_url(
                self.client_config) + IWUtils.get_query_params_string_from_dict(params=params)
            workflow_run_tasks_list = []
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_workflow_run_tasks,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                initial_msg = response.get("message", "")
                result = response.get("result", {}).get("items", [])
                while len(result) > 0:
                    workflow_run_tasks_list.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", {}).get("items", [])
            response["result"] = workflow_run_tasks_list
            response["message"] = initial_msg
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to get workflow run task details.')
            raise WorkflowError('Error occurred while trying to get workflow run task details.')

    def get_workflow_run_details(self, workflow_id=None, run_id=None):
        """
        Gets Infoworks Data workflow run details for given workflow id
        :param workflow_id: id of the workflow whose details are to be fetched
        :type workflow_id: String
        :param run_id: Workflow run id
        :type run_id: String
        :return: response dict
        """
        if None in {run_id, workflow_id}:
            self.logger.error("run_id id or workflow_id cannot be None")
            raise Exception("run_id id or workflow_id cannot be None")
        response = None
        try:
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.get_workflow_run_id_details_url(self.client_config, workflow_id, run_id)
                                                               , IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            result = response.get('result', {})
            if result.get('id', None) is None:
                self.logger.error('Failed to find the workflow run details')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_code=ErrorCode.USER_ERROR,
                                                     error_desc='Failed to find the workflow run details',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully got the workflow run details.')
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to get workflow run details.')
            raise WorkflowError('Error occurred while trying to get workflow run details.')

    def get_workflow_lineage(self, domain_id, workflow_id):
        """
        Gets workflow lineage
        :param domain_id: Domain ID of which Workflow belongs
        :type domain_id: String
        :param workflow_id: Entity id of the Workflow
        :type workflow_id: String
        :return: Response Dict
        """
        if domain_id is None or workflow_id is None:
            self.logger.error("domain_id or workflow_id cannot be None")
            raise Exception("domain_id or workflow_id cannot be None")

        response = None

        try:
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.workflow_lineage_url(
                self.client_config, domain_id, workflow_id), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)

            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to get workflow lineage.')
            raise WorkflowError('Error occurred while trying to get workflow lineage.')


import os

from infoworks.sdk.cicd.download_configurations.get_pipeline_configuration import DownloadPipeline
from infoworks.sdk.cicd.download_configurations.get_workflow_configuration import DownloadWorkflow
from infoworks.sdk.url_builder import create_pipeline_url, create_workflow_url
from infoworks.sdk.base_client import BaseClient
from infoworks.sdk.utils import IWUtils


class DownloadAllEntitiesFromDomain(BaseClient):
    def __init__(self):
        super(DownloadAllEntitiesFromDomain, self).__init__()

    def callurl(self, nextUrl):
        try:
            self.logger.info("url {url}".format(url=nextUrl))
            response = self.call_api("GET", nextUrl,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))

            parsed_response = IWUtils.ejson_deserialize(response.content)
            if parsed_response is not None and response.status_code == 200:
                return parsed_response
        except Exception as e:
            raise Exception("Unable to get response for url: {url}".format(url=nextUrl))

    def cicd_get_all_configuration_dumps_from_domain(self, domain_ids, config_file_dump_path, files_overwrite=True,
                                                     serviceaccountemail="admin@infoworks.io",
                                                     replace_words=""):
        if not os.path.exists(os.path.join(config_file_dump_path, "modified_files")):
            os.makedirs(os.path.join(config_file_dump_path, "modified_files"))
        if not os.path.exists(os.path.join(config_file_dump_path, "source")):
            os.makedirs(os.path.join(config_file_dump_path, "source"))
        if not os.path.exists(os.path.join(config_file_dump_path, "pipeline")):
            os.makedirs(os.path.join(config_file_dump_path, "pipeline"))
        if not os.path.exists(os.path.join(config_file_dump_path, "workflow")):
            os.makedirs(os.path.join(config_file_dump_path, "workflow"))
        for domain_id in domain_ids:
            get_pipeline_url = create_pipeline_url(self.client_config, str(domain_id))
            pipeline_ids = []
            response = self.call_api("GET", get_pipeline_url,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))

            parsed_response = IWUtils.ejson_deserialize(response.content)
            if response.status_code == 200:
                result = parsed_response.get("result", [])
                while len(result) > 0:
                    pipeline_ids.extend([item["id"] for item in result])
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.json().get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = self.callurl(nextUrl)
                    result = response.get("result", [])
            get_workflow_url = create_workflow_url(self.client_config, str(domain_id))
            workflow_ids = []
            response = self.call_api("GET", get_workflow_url,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))

            parsed_response = IWUtils.ejson_deserialize(response.content)
            if response.status_code == 200:
                result = parsed_response.get("result", [])
                while len(result) > 0:
                    workflow_ids.extend([item["id"] for item in result])
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.json().get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = self.callurl(nextUrl)
                    result = response.get("result", [])

            if len(pipeline_ids) > 0:
                pipeline_obj = DownloadPipeline()
                pipeline_obj.cicd_get_pipelineconfig_dumps(pipeline_ids, config_file_dump_path, files_overwrite,
                                                           serviceaccountemail,
                                                           replace_words)
            else:
                print("No pipelines available to dump the configurations ")
            if len(workflow_ids) > 0:
                wf_obj = DownloadWorkflow()
                wf_obj.cicd_get_workflowconfig_dumps(workflow_ids, config_file_dump_path, files_overwrite,
                                                     serviceaccountemail,
                                                     replace_words)
            else:
                print("No workflow available to dump the configurations ")



POLLING_FREQUENCY_IN_SEC = 15
NUM_POLLING_RETRIES = 3
POLLING_TIMEOUT = 1200
NUM_FAILURE_RETRIES = 5
RUNNING_JOB_LOGGING_COUNTER = 10
REQUEST_TIMEOUT_IN_SEC = 60
MAX_RETRIES = 1
LOG_LOCATION = "/tmp/iwx_sdk.log"


class Response(object):
    class Code(object):
        OK = 200

    class Status(object):
        FAILED = 'failed'
        SUCCESS = 'success'
        ERROR = 'error'
        PENDING = 'pending'
        RUNNING = 'running'


class EntityType(object):
    TABLE = 'table'
    SOURCE = 'source'
    PIPELINE = 'pipeline'
    PIPELINE_TARGET = 'pipeline_target'
    JOB = 'job'
    CONFIG = 'config'
    TABLE_GROUP = 'table_group'
    WORKFLOW = 'workflow'


class Job(object):
    class Status(object):
        PENDING = 'pending'
        RUNNING = 'running'
        BLOCKED = 'blocked'
        FAILED = 'failed'
        COMPLETED = 'completed'
        CANCELED = 'canceled'
        ABORTED = 'aborted'
        SUCCESS = 'success'


class ErrorCode(object):
    POLL_TIMEOUT = 1408
    ENTITY_NOT_FOUND = 1400
    DEPENDENT_ENTITY_NOT_FOUND = 1401
    ENTITY_STATE_INCONSISTENT = 1402
    DEPENDENT_ENTITY_STATE_INCONSISTENT = 1403
    ACTION_LOCKED_FOR_ENTITY = 1404
    MANDATORY_PARAMETER_MISSING = 1405
    ENTITY_LOCKED = 1500
    DEPENDENT_ENTITY_LOCKED = 1501
    COMMUNICATION_FAILURE = 1600
    FSM_ERROR = 1601
    INTERNAL_SERVER_ERROR = 500
    RESOURCE_NOT_FOUND = 404
    USER_ERROR = 1406
    GENERIC_ERROR = 1407


class SourceMappings(object):
    oracle = {"sub_type": "oracle", "driver_name": "oracle.jdbc.driver.OracleDriver", "driver_version": "v2",
              "database": "ORACLE"}
    mysql = {"sub_type": "mysql", "driver_name": "org.mariadb.jdbc.Driver", "driver_version": "v2",
              "database": "MYSQL"}
    sqlserver = {"sub_type": "sqlserver", "driver_name": "com.microsoft.sqlserver.jdbc.SQLServerDriver", "driver_version": "v2",
              "database": "SQL Server"}
    teradata = {"sub_type": "teradata", "driver_name": "com.teradata.jdbc.TeraDriver", "driver_version": "v2",
                "database": "TERADATA"}
    netezza = {"sub_type": "netezza", "driver_name": "org.netezza.Driver", "driver_version": "v2",
               "database": "NETEZZA"}




from infoworks.sdk.cicd.download_configurations.get_source_configuration import DownloadSource
from infoworks.sdk.cicd.download_configurations.get_pipeline_configuration import DownloadPipeline
from infoworks.sdk.cicd.download_configurations.get_workflow_configuration import DownloadWorkflow
from infoworks.sdk.cicd.download_configurations.get_entity_configuration_lineage import DownloadEntityWithLineage
from infoworks.sdk.cicd.download_configurations.get_all_configs_from_domain import DownloadAllEntitiesFromDomain
from infoworks.sdk.cicd.upload_configurations.wrappersource import WrapperSource
from infoworks.sdk.cicd.upload_configurations.wrapperpipeline import WrapperPipeline
from infoworks.sdk.cicd.upload_configurations.wrapperworkflow import WrapperWorkflow
from infoworks.sdk.source_client import SourceClient
from infoworks.sdk.pipeline_client import PipelineClient
from infoworks.sdk.workflow_client import WorkflowClient
from infoworks.sdk.domain_client import DomainClient
from infoworks.sdk.jobmetrics import JobMetricsClient
from infoworks.sdk.admin_client import AdminClient
from infoworks.sdk.jobs_client import JobsClient
from infoworks.sdk.pipeline_group_client import PipelineGroupClient
from infoworks.sdk.replicator_client import ReplicatorClient


class InfoworksClientSDK(SourceClient, PipelineClient, WorkflowClient, DomainClient, AdminClient, JobMetricsClient,
                         DownloadSource, DownloadPipeline, DownloadWorkflow, DownloadEntityWithLineage,
                         DownloadAllEntitiesFromDomain, WrapperSource, WrapperPipeline, WrapperWorkflow, JobsClient,
                         PipelineGroupClient, ReplicatorClient):
    def __init__(self):
        super().__init__()



import copy
import datetime
import math
import traceback
from collections import OrderedDict
from infoworks.sdk import url_builder
from infoworks.sdk.utils import IWUtils
from infoworks.error import AdminError
from infoworks.sdk.base_client import BaseClient
import pandas as pd
from concurrent.futures import ThreadPoolExecutor


class JobMetricsClient(BaseClient):
    job_type_mappings = {
        'EXPORT_DATA': 'EXPORT',
        'SOURCE_CRAWL': 'FULL_LOAD',
        'SOURCE_CDC': 'INCREMENTAL',
        'SOURCE_CDC_MERGE': 'FULL_LOAD',
        'SOURCE_STRUCTURED_CRAWL': 'FULL_LOAD',
        'FULL_LOAD': 'FULL_LOAD',
        'CDC': 'INCREMENTAL',
        'MERGE': 'INCREMENTAL',
        'SOURCE_MERGE': 'INCREMENTAL',
        'SOURCE_STRUCTURED_CDC_MERGE': 'INCREMENTAL',
        'SOURCE_STRUCTURED_CDC': 'INCREMENTAL',
        'SOURCE_SEMISTRUCTURED_CDC_MERGE': 'INCREMENTAL',
        'SOURCE_SEMISTRUCTURED_CDC': 'INCREMENTAL'
    }

    def __init__(self):
        super(JobMetricsClient, self).__init__()
        self.job_metrics_final = []

    def get_tablegroup_id_from_job(self, job_id, source_id):
        try:
            url_to_get_jobinfo = url_builder.submit_source_job(self.client_config, source_id) + f"/{job_id}"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_jobinfo,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result")
                table_group_id = result.get("sub_entity_id", "")
                processedAt = result.get("processed_at", "")
                job_end_time = result.get("last_updated", "")
                status = result.get("status", "")
                entity_type = result.get("entity_type", "")
                return table_group_id, processedAt, job_end_time, status, entity_type
        except Exception as e:
            raise AdminError("Unable to get table group details from the Job Id")

    def get_table_group_name(self, table_group_id, source_id):
        try:
            url_to_get_tginfo = url_builder.create_table_group_url(self.client_config, source_id) + f"/{table_group_id}"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_tginfo,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", {})
                return result.get('name'), result.get('tables')
        except Exception as e:
            raise AdminError("Unable to get table group info")

    def get_ingestion_metrics(self, job_id, source_id):
        """
        Function to get ingestion metrics for the job
        :param job_id: Entity identifier of the job
        :param source_id: Entity identifier of the source
        :return: List of job metrics
        """
        try:
            combinedJobMetric = []
            url_to_get_job_ingestion_metrics = url_builder.get_job_metrics_url(self.client_config, source_id,
                                                                               job_id) + "?limit=50&offset=0"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_job_ingestion_metrics,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", [])
                while len(result) > 0:
                    combinedJobMetric.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(self.callurl(nextUrl).content)
                    result = response.get("result", [])
                return combinedJobMetric
        except Exception as e:
            raise AdminError("Unable to get ingestion job metrics info")

    def get_export_metrics(self, job_id, source_id):
        """
        Function to get export metrics for the source job
        :param job_id: Entity identifier of the job
        :param source_id: Entity identifier of the source
        :return: List of job metrics
        """
        try:
            combinexportMetric = []
            url_to_get_export_job_metrics = url_builder.get_export_metrics_source_url(self.client_config, source_id,
                                                                                      job_id) + "?limit=50&offset=0"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_export_job_metrics,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", [])
                while len(result) > 0:
                    combinexportMetric.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(self.callurl(nextUrl).content)
                    result = response.get("result", [])
                return combinexportMetric
        except Exception as e:
            raise AdminError("Unable to get export job metrics info")

    def get_ingestion_metrics_with_job_id(self, job_id):
        """
        Function to get ingestion metrics for the job
        :param job_id: Entity identifier of the job
        :return: List of job metrics
        """
        try:
            combinedJobMetric = []
            url_to_get_job_ingestion_metrics = url_builder.get_ingestion_metrics_admin_url(self.client_config,
                                                                                           job_id) + "?limit=50&offset=0"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_job_ingestion_metrics,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", [])
                while len(result) > 0:
                    combinedJobMetric.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(self.callurl(nextUrl).content)
                    result = response.get("result", [])
                return combinedJobMetric
        except Exception as e:
            raise AdminError("Unable to get ingestion job metrics info")

    def get_export_metrics_with_job_id(self, job_id):
        """
        Function to get ingestion metrics for the job
        :param job_id: Entity identifier of the job
        :return: List of job metrics
        """
        try:
            exportMetric = []
            url_to_get_job_ingestion_metrics = url_builder.get_export_metrics_url(self.client_config,
                                                                                  job_id) + "?limit=50&offset=0"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_job_ingestion_metrics,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", [])
                while len(result) > 0:
                    exportMetric.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(self.callurl(nextUrl).content)
                    result = response.get("result", [])
                return exportMetric
        except Exception as e:
            raise AdminError("Unable to get export job metrics info")

    def get_metrics_prodops(self, config_body):
        """
        Function to Fetch the metrics for a given time period
        :param config_body: JSON dict
        config_body = {'limit': 10000, 'date_range': {'type': 'last', 'unit': 'day', 'value': 1}, 'offset': 0}
        :return: List of job metrics
        """
        try:
            if config_body is None:
                raise Exception("config_body cannot be None")
            metrics = []
            url_to_get_metrics = url_builder.get_metrics_prodops_url(self.client_config) + "?limit=50&offset=0"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_metrics,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", [])
                while len(result) > 0:
                    metrics.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(self.callurl(nextUrl).content)
                    result = response.get("result", [])
                return metrics
        except Exception as e:
            raise AdminError("Unable to get job metrics info")

    def get_source_file_paths(self, source_id, table_id, job_id):
        try:
            source_files = []
            url_to_get_src_filepaths = url_builder.get_source_file_paths_url(self.client_config, source_id, table_id,
                                                                             job_id) + "?limit=50&offset=0"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_src_filepaths,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", [])
                while len(result) > 0:
                    source_files.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(self.callurl(nextUrl).content)
                    result = response.get("result", [])
                return source_files
        except Exception as e:
            print(e)
            raise AdminError("Unable to get source file pat metrics info")

    def get_pipeline_build_metrics(self, job_id):
        try:
            combinedJobMetric = []
            url_to_get_pipeline_build_metrics = url_builder.get_pipeline_build_metrics_url(self.client_config,
                                                                                           job_id) + "?limit=50&offset=0"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_pipeline_build_metrics,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", [])
                while len(result) > 0:
                    combinedJobMetric.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(self.callurl(nextUrl).content)
                    result = response.get("result", [])
                return combinedJobMetric
        except Exception as e:
            raise AdminError("Unable to get pipeline job metrics info")

    def get_source_info(self):
        try:
            combined_sources = []
            url_to_get_get_source_info = url_builder.get_source_details_url(
                self.client_config) + "?limit=50&offset=0"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_get_source_info,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", [])
                while len(result) > 0:
                    combined_sources.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(self.callurl(nextUrl).content)
                    result = response.get("result", [])
                return combined_sources
        except Exception as e:
            raise AdminError("Unable to get source details")

    def get_jobs_of_single_source(self, source_id, date_string):
        try:
            combinedJobs = []
            # filter_condition = "{\"$and\": [{\"createdAt\": {\"$gte\": {\"$date\": \"" + date_string + "\"}}},{\"jobType\": {\"$in\": [\"source_crawl\", \"source_cdc_merge\"]}}]}"
            filter_condition = "{\"$and\": [{\"entityId\":\"" + source_id + "\"},{\"last_upd\": {\"$gte\": {\"$date\": \"" + date_string + "\"}}},{\"jobType\": {\"$in\": [\"source_crawl\",\"source_structured_crawl\",\"source_cdc\",\"source_cdc_merge\",\"export_data\",\"full_load\",\"cdc\",\"source_merge\",\"source_structured_cdc_merge\",\"source_structured_cdc\",\"source_cdc_merge\",\"source_semistructured_cdc_merge\",\"source_semistructured_cdc\"]}}, {\"status\":{\"$in\":[\"failed\",\"completed\",\"running\",\"pending\",\"canceled\",\"aborted\"]}}]}"
            url_to_get_jobs = url_builder.get_job_status_url(
                self.client_config) + f"?filter={filter_condition}&limit=50&offset=0"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_jobs,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)

            if response is not None and "result" in response:
                result = response.get("result", [])
                while len(result) > 0:
                    combinedJobs.extend(result)
                    next_url = response.get('links')['next']
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=next_url,
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(self.callurl(nextUrl).content)
                    result = response.get("result", [])
                return combinedJobs
        except Exception as e:
            raise AdminError("Unable to get ingestion jobs list of source")

    def get_pipeline_jobs(self, date_string):
        try:
            combinedJobs = []
            filter_condition = "{\"$and\": [{\"last_upd\": {\"$gte\": {\"$date\": \"" + date_string + "\"}}},{\"jobType\": {\"$in\": [\"pipeline_build\"]}}, {\"status\":{\"$in\":[\"failed\",\"completed\",\"running\",\"pending\",\"aborted\",\"canceled\"]}}]}"
            url_to_get_jobs = url_builder.get_job_status_url(
                self.client_config) + f"?filter={filter_condition}&limit=50&offset=0"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_jobs,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)

            if response is not None and "result" in response:
                result = response.get("result", [])
                while len(result) > 0:
                    combinedJobs.extend(result)
                    next_url = response.get('links')['next']
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=next_url,
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(self.callurl(nextUrl).content)
                    result = response.get("result", [])
                return combinedJobs
        except Exception as e:
            raise AdminError("Unable to get pipeline jobs list")

    def get_all_source_jobs(self, date_string):
        try:
            combinedJobs = []
            filter_condition = "{\"$and\": [{\"last_upd\": {\"$gte\": {\"$date\": \"" + date_string + "\"}}},{\"entityType\": {\"$in\": [\"source\"]}}, {\"jobType\": {\"$in\": [\"source_crawl\",\"source_structured_crawl\",\"source_cdc\",\"source_cdc_merge\",\"export_data\",\"full_load\",\"cdc\",\"source_merge\",\"source_structured_cdc_merge\",\"source_structured_cdc\",\"source_cdc_merge\",\"source_semistructured_cdc_merge\",\"source_semistructured_cdc\"]}},{\"status\":{\"$in\":[\"failed\",\"completed\",\"running\",\"pending\",\"canceled\",\"aborted\"]}}]}"
            url_to_get_all_source_jobs = url_builder.get_job_status_url(
                self.client_config) + f"?filter={filter_condition}&limit=50&offset=0"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_all_source_jobs,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", [])
                while len(result) > 0:
                    combinedJobs.extend(result)
                    next_url = response.get('links')['next']
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=next_url,
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(self.callurl(nextUrl).content)
                    result = response.get("result", [])
                return combinedJobs
        except Exception as e:
            raise AdminError("Unable to get source ingestion jobs list")

    def callurl(self, url):
        try:
            response = self.call_api("GET", url,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                                     )
            if response is not None:
                return response
        except Exception as e:
            raise AdminError("Unable to get response for url: {url}".format(url=url))

    def get_table_info(self, source_id, table_id):
        try:
            url_to_get_table_info = url_builder.get_table_configuration(
                self.client_config, source_id, table_id)
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_table_info,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None:
                result = response.get("result", {})
                return result
        except Exception as e:
            raise AdminError("Unable to get table info")

    def get_table_export_info(self, source_id, table_id):
        try:
            url_to_get_table_info = url_builder.table_export_config_url(
                self.client_config, source_id, table_id)
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_table_info,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", {})
                return result
        except Exception as e:
            raise AdminError("Unable to get export job metrics info")

    def get_pipeline_name(self, domain_id, pipeline_id):
        try:
            url_to_get_pipeline_info = url_builder.get_pipeline_url(self.client_config, domain_id, pipeline_id)
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_pipeline_info,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", None)
                return result.get('name')
        except:
            raise AdminError("Unable to get pipeline NAME")

    def get_source_jobs_metrics_results(self, date_string, source, workflow_id=None, workflow_run_id=None):
        """
        Gets the Infoworks source job metrics
        :param date_string: date string to fetch the data till that date
        :type date_string: String
        :param workflow_id: Workflow id to get the jobs
        :type workflow_id: String
        :param workflow_run_id: Workflow run id to get the jobs
        :type workflow_run_id: String
        :return: response dict
        """
        src_name = source["name"]
        list_of_jobs_obj = self.get_jobs_of_single_source(source["id"], date_string)
        try:
            for job in list_of_jobs_obj:
                job_id = job["id"]
                job_type = job["type"]
                job_status = job["status"]
                job_createdAt = job["created_at"]
                running_job_template = {
                    'workflow_id': job.get('triggered_by', {}).get('parent_id', ''),
                    'workflow_run_id': job.get('triggered_by', {}).get('run_id', ''),
                    'job_id': job_id,
                    'job_type': self.job_type_mappings[job_type.upper()],
                    'job_start_time': job_createdAt.split('.')[0].replace('T', ' '),
                    'job_end_time': "",
                    "source_file_names": [],
                    'job_status': job_status.upper(),
                    'entity_type': "source",
                    "source_name": src_name, "source_schema_name": job.get("source_schema_name", ""),
                    "source_database_name": job.get("source_database_name", ""), "table_group_name": "",
                    "iwx_table_name": "", 'starting_watermark_value': '', 'ending_watermark_value': '',
                    "pre_target_count": "", "fetch_records_count": 0,
                    "target_records_count": ""}
                running_od = OrderedDict()
                if job_status.upper() in ["RUNNING", "PENDING", "CANCELED", "ABORTED", "FAILED"]:
                    for key in ['workflow_id', 'workflow_run_id', 'job_id', 'entity_type', 'job_type', 'job_start_time',
                                'job_end_time', 'job_status',
                                'source_name', 'source_file_names', 'source_schema_name', 'source_database_name',
                                'table_group_name',
                                'iwx_table_name', 'starting_watermark_value', 'ending_watermark_value',
                                'target_schema_name', 'target_table_name', 'pre_target_count', 'fetch_records_count',
                                'target_records_count']:
                        running_od[key] = running_job_template.get(key, "")
                    self.job_metrics_final.append(running_od)
                    continue
                try:
                    tg_id, processedAt, job_end_time, job_status, entity_type = self.get_tablegroup_id_from_job(
                        job_id,
                        source["id"])
                    table_group_name, all_tables_list = self.get_table_group_name(tg_id, source["id"])
                except:
                    # This means the job is non-tablegroup job
                    table_group_name = ""
                    entity_type = "TABLE"
                    all_tables_list = []
                ing_metrics = self.get_ingestion_metrics(str(job_id), source["id"])
                successful_tables = []
                if ing_metrics != [] and job_type != "export_data":
                    tables_list = list(set([i['table_id'] for i in ing_metrics]))
                    df = pd.DataFrame(ing_metrics)
                    for table_id in tables_list:
                        successful_tables.append(table_id)
                        filter1 = df["table_id"] == table_id
                        table = {}
                        if len(df.loc[filter1]) > 1:
                            # Incremental Job
                            filter2 = df["job_type"] == "CDC"
                            filter3 = df["job_type"] == "MERGE"
                            cdc_output = df.loc[filter1 & filter2].to_dict('records')[0]
                            merge_output = df.loc[filter1 & filter3].to_dict('records')[0]
                            for item in ['source_id', 'fetch_records_count', 'job_id', 'job_start_time',
                                         'source_schema_name', 'source_database_name']:
                                table[item] = cdc_output.get(item, "")
                                table['job_type'] = "INCREMENTAL"
                            for item in ['workflow_id', 'workflow_run_id', 'job_end_time', 'job_status',
                                         'target_records_count', 'first_merged_watermark', 'last_merged_watermark']:
                                table[item] = merge_output.get(item, "")
                                table['job_type'] = "INCREMENTAL"
                        else:
                            table = df.loc[filter1].to_dict('records')[0]
                        table_name_url = self.get_table_info(table.get('source_id'), table_id)
                        table_name = table_name_url.get('name')
                        target_table_name = table_name_url.get("configuration", {}).get('target_table_name', '')
                        target_schema_name = table_name_url.get("configuration", {}).get('target_schema_name', '')
                        table["source_name"] = src_name
                        table["source_schema_name"] = table.get("source_schema_name", "")
                        table["source_database_name"] = table.get("source_database_name", "")
                        table["source_file_names"] = self.get_source_file_paths(source['id'], table_id,
                                                                                job_id) if table[
                                                                                               "source_schema_name"] == "" else []
                        table["table_group_name"] = table_group_name
                        table["iwx_table_name"] = table_name
                        table["target_schema_name"] = target_schema_name
                        table["target_table_name"] = target_table_name
                        table["workflow_id"] = table.get("workflow_id", "")
                        table["workflow_run_id"] = table.get('workflow_run_id', "")
                        if workflow_id is not None and workflow_run_id is not None:
                            if not (table["workflow_id"] == workflow_id and table[
                                "workflow_run_id"] == workflow_run_id):
                                continue
                        table["entity_type"] = entity_type
                        table["starting_watermark_value"] = table.pop('first_merged_watermark', '')
                        table["ending_watermark_value"] = table.pop('last_merged_watermark', '')
                        if math.isnan(table.get('target_records_count')):
                            table['pre_target_count'] = None
                            table['target_records_count'] = None
                        else:
                            if table["job_status"] == "FAILED":
                                table['pre_target_count'] = table.get('target_records_count')
                                table['target_records_count'] = int(table.get('target_records_count'))
                            else:
                                table['pre_target_count'] = int(
                                    table.get('target_records_count') - int(table.get('fetch_records_count')))
                                table['target_records_count'] = int(table.get('target_records_count'))
                        if table.get('job_type') == "CDC":
                            table['job_type'] = "INCREMENTAL"
                        od = OrderedDict()
                        table['job_start_time'] = table['job_start_time'].split('.')[0].replace('T', ' ')
                        table['job_end_time'] = table['job_end_time'].split('.')[0].replace('T', ' ')
                        table['fetch_records_count'] = int(table['fetch_records_count'])
                        for key in ['workflow_id', 'workflow_run_id', 'job_id', 'entity_type', 'job_type',
                                    'job_start_time',
                                    'job_end_time', 'job_status',
                                    'source_name', 'source_file_names', 'source_schema_name', 'source_database_name',
                                    'table_group_name',
                                    'iwx_table_name', 'starting_watermark_value', 'ending_watermark_value',
                                    'target_schema_name', 'target_table_name', 'pre_target_count',
                                    'fetch_records_count',
                                    'target_records_count']:
                            od[key] = table.get(key, '')
                        self.job_metrics_final.append(od)

                for table in all_tables_list:
                    if table['table_id'] not in successful_tables and job_type != "export_data":
                        tableInfo = self.get_table_info(source["id"], table['table_id'])
                        table_name = tableInfo.get("name")
                        row_count = tableInfo.get('row_count', 0)
                        synctype = tableInfo.get('configuration').get('sync_type')
                        temp_failed = OrderedDict()
                        temp = {
                            'workflow_id': job.get('triggered_by', {}).get('parent_id', ''),
                            'workflow_run_id': job.get('triggered_by', {}).get('run_id', ''),
                            'job_id': job_id,
                            'job_type': synctype.upper(),
                            'job_start_time': processedAt.split('.')[0].replace('T', ' '),
                            'job_end_time': job_end_time.split('.')[0].replace('T', ' '),
                            'job_status': job_status.upper(),
                            'entity_type': entity_type,
                            "source_file_names": [],
                            "target_table_name": tableInfo.get("target_table_name", ""),
                            "target_schema_name": tableInfo.get("target_schema_name", ""),
                            "source_name": src_name, "source_schema_name": job.get("source_schema_name", ""),
                            "source_database_name": job.get("source_database_name", ""),
                            "table_group_name": table_group_name, "iwx_table_name": table_name,
                            'starting_watermark_value': '', 'ending_watermark_value': '',
                            "pre_target_count": row_count, "fetch_records_count": 0,
                            "target_records_count": row_count}
                        for key in ['workflow_id', 'workflow_run_id', 'job_id', 'entity_type', 'job_type',
                                    'job_start_time',
                                    'job_end_time', 'job_status',
                                    'source_name', 'source_file_names', 'source_schema_name', 'source_database_name',
                                    'table_group_name',
                                    'iwx_table_name', 'starting_watermark_value', 'ending_watermark_value',
                                    'target_schema_name', 'target_table_name', 'pre_target_count',
                                    'fetch_records_count',
                                    'target_records_count']:
                            temp_failed[key] = temp.get(key, "")

                        self.job_metrics_final.append(temp_failed)

                if job_type == "export_data":
                    export_metrics = self.get_export_metrics(str(job_id))
                    export_successful_tables = []
                    if export_metrics is not None:
                        tables_list = list([i['table_id'] for i in export_metrics])
                        df_export = pd.DataFrame(export_metrics)
                        for table_id in tables_list:
                            export_successful_tables.append(table_id)
                            filter1 = df_export["table_id"] == table_id
                            table = df_export.loc[filter1].to_dict('records')[0]
                            table["job_type"] = "EXPORT"
                            table_name_url = self.get_table_info(table.get('source_id'), table_id)
                            table_export_config = self.get_table_export_info(table.get('source_id'), table_id)
                            table_name = table_name_url.get('name')
                            table["source_name"] = src_name
                            table["table_group_name"] = table_group_name
                            table["iwx_table_name"] = table_name
                            if table_export_config.get("target_type", "") == "BIGQUERY":
                                table["target_schema_name"] = ".".join(
                                    [table_export_config.get("connection", {}).get("project_id", ""),
                                     table_export_config.get("target_configuration", {}).get("dataset_name", "")])
                            else:
                                table["target_schema_name"] = ".".join(
                                    [table_export_config.get("target_configuration", {}).get("schema_name", ""),
                                     table_export_config.get("target_configuration", {}).get("database_name", "")])
                            table["target_table_name"] = table_export_config.get("target_configuration", {}).get(
                                "table_name", "")
                            table["workflow_id"] = table.get("workflow_id", "")
                            table["workflow_run_id"] = table.get("workflow_run_id", "")
                            table["entity_type"] = entity_type
                            table["source_file_names"] = []
                            table["starting_watermark_value"] = table.pop('first_merged_watermark', '')
                            table["ending_watermark_value"] = table.pop('last_merged_watermark', '')
                            if math.isnan(table.get('target_records_count')):
                                table['pre_target_count'] = None
                                table['target_records_count'] = None
                            else:
                                if table["job_status"] == "FAILED":
                                    table['pre_target_count'] = table.get('target_records_count')
                                    table['target_records_count'] = int(table.get('target_records_count'))
                                else:
                                    table['pre_target_count'] = int(
                                        table.get('target_records_count') - int(table.get('number_of_records_written')))
                                    table['target_records_count'] = int(table.get('target_records_count'))
                            od = OrderedDict()
                            table['job_start_time'] = table['job_start_time'].split('.')[0].replace('T', ' ')
                            table['job_end_time'] = table['job_end_time'].split('.')[0].replace('T', ' ')
                            table['fetch_records_count'] = int(table['number_of_records_written'])
                            for key in ['workflow_id', 'workflow_run_id', 'job_id', 'entity_type', 'job_type',
                                        'job_start_time', 'job_end_time', 'job_status',
                                        'source_name', 'source_file_names', 'source_schema_name',
                                        'source_database_name',
                                        'table_group_name',
                                        'iwx_table_name', 'starting_watermark_value', 'ending_watermark_value',
                                        'target_schema_name', 'target_table_name', 'pre_target_count',
                                        'fetch_records_count',
                                        'target_records_count']:
                                od[key] = table.get(key, "")
                            self.job_metrics_final.append(od)

                    for table in all_tables_list:
                        if table['table_id'] not in export_successful_tables:
                            tableInfo = self.get_table_info(source["id"], table['table_id'])
                            table_name = tableInfo.get("name")
                            row_count = tableInfo.get('row_count', 0)
                            export_failed_od = OrderedDict()
                            temp = {
                                'workflow_id': job.get('triggered_by', {}).get('parent_id', ''),
                                'workflow_run_id': job.get('triggered_by', {}).get('run_id', ''),
                                'job_id': job_id,
                                'job_type': "EXPORT",
                                'job_start_time': processedAt.split('.')[0].replace('T', ' '),
                                'job_end_time': job_end_time.split('.')[0].replace('T', ' '),
                                'job_status': job_status.upper(),
                                'entity_type': entity_type,
                                "source_file_names": [],
                                "source_name": src_name, "table_group_name": table_group_name,
                                "iwx_table_name": table_name,
                                'starting_watermark_value': '', 'ending_watermark_value': '',
                                "pre_target_count": row_count, "fetch_records_count": 0,
                                "target_records_count": row_count}
                            for key in ['workflow_id', 'workflow_run_id', 'job_id', 'entity_type', 'job_type',
                                        'job_start_time', 'job_end_time', 'job_status',
                                        'source_name', 'source_file_names', 'source_schema_name',
                                        'source_database_name',
                                        'table_group_name',
                                        'iwx_table_name', 'starting_watermark_value', 'ending_watermark_value',
                                        'target_schema_name', 'target_table_name', 'pre_target_count',
                                        'fetch_records_count',
                                        'target_records_count']:
                                export_failed_od[key] = temp.get(key, "")
                            self.job_metrics_final.append(export_failed_od)
        except Exception as e:
            print(str(e))
            traceback.print_exc()

    def get_pipeline_build_metrics_results(self, job=None, workflow_id=None, workflow_run_id=None):
        """
        Gets the Infoworks pipeline build metrics
        :param job: job object to get the pipeline build metrics details
        :type job: JSON Object
        :param workflow_id: Workflow id to get the jobs
        :type workflow_id: String
        :param workflow_run_id: Workflow run id to get the jobs
        :type workflow_run_id: String
        :return: response dict
        """
        if job is None:
            self.logger.error("job is a mandatory parameter")
            raise Exception("job is a mandatory parameter")
        temp = []
        job_id = job['id']
        job_type = job['type']
        job_status = job['status']
        job_createdAt = job['created_at']
        entity_type = job['entity_type']
        running_job_template = {
            'workflow_id': job.get('triggered_by', {}).get('parent_id', ''),
            'workflow_run_id': job.get('triggered_by', {}).get('run_id', ''),
            'job_id': job_id,
            'job_type': job_type.upper(),
            'job_start_time': job_createdAt.split('.')[0].replace('T', ' '),
            'job_end_time': "",
            "source_file_names": [],
            'job_status': job_status.upper(),
            'entity_type': entity_type,
            "source_name": "", "source_schema_name": job.get("source_schema_name", ""),
            "source_database_name": job.get("source_database_name", ""), "table_group_name": "", "iwx_table_name": "",
            'starting_watermark_value': '', 'ending_watermark_value': '',
            "pre_target_count": "", "fetch_records_count": 0,
            "target_records_count": ""}
        running_od = OrderedDict()
        if job_status.upper() in ["RUNNING", "PENDING", "CANCELED", "ABORTED", "FAILED"]:
            for key in ['workflow_id', 'workflow_run_id', 'job_id', 'entity_type', 'job_type', 'job_start_time',
                        'job_end_time', 'job_status',
                        'source_name', 'source_file_names', 'source_schema_name', 'source_database_name',
                        'table_group_name',
                        'iwx_table_name', 'starting_watermark_value', 'ending_watermark_value', 'target_schema_name',
                        'target_table_name', 'pre_target_count', 'fetch_records_count',
                        'target_records_count']:
                running_od[key] = running_job_template.get(key, "")
            self.job_metrics_final.append(running_od)
            return
        pipeline_metrics = self.get_pipeline_build_metrics(str(job_id))
        pipeline_successful_tables = []
        if pipeline_metrics is not None:
            tables_list = list(set([i['target_table_name'] for i in pipeline_metrics]))
            df_pipeline = pd.DataFrame(pipeline_metrics)
            for target_table_name in tables_list:
                schema_name, table_name = target_table_name.split(".")
                pipeline_successful_tables.append(table_name)
                filter1 = df_pipeline["target_table_name"] == target_table_name
                table = df_pipeline.loc[filter1].to_dict('records')[0]
                table["job_type"] = "PIPELINE_BUILD"
                table["source_name"] = ''
                table["source_file_names"] = []
                table["source_schema_name"] = ""
                table["source_database_name"] = ""
                table["table_group_name"] = ''
                table["iwx_table_name"] = ''
                table['target_schema_name'] = schema_name
                table['target_table_name'] = table_name
                table["workflow_id"] = table.get('workflow_id', '')
                table["workflow_run_id"] = table.get('workflow_run_id', '')
                if workflow_id is not None and workflow_run_id is not None:
                    if not (table["workflow_id"] == workflow_id and table[
                        "workflow_run_id"] == workflow_run_id):
                        continue
                table["entity_type"] = job.get("entity_type", "pipeline")
                table["starting_watermark_value"] = table.pop('first_merged_watermark', '')
                table["ending_watermark_value"] = table.pop('last_merged_watermark', '')
                table["target_records_count"] = table.get("target_records_count", None)
                table["job_status"] = table.get("job_status", "")
                if math.isnan(table.get('target_records_count')):
                    table['pre_target_count'] = None
                    table['target_records_count'] = None
                else:
                    if table["job_status"] == "FAILED":
                        table['pre_target_count'] = table.get('target_records_count')
                        table['target_records_count'] = int(table.get('target_records_count'))
                    else:
                        table['pre_target_count'] = int(
                            table.get('target_records_count') - int(table.get('number_of_records_written')))
                        table['target_records_count'] = int(table.get('target_records_count'))
                    od = OrderedDict()
                    table['job_start_time'] = table['job_start_time'].split('.')[0].replace('T', ' ')
                    table['job_end_time'] = table['job_end_time'].split('.')[0].replace('T', ' ')
                    table['fetch_records_count'] = int(table['number_of_records_written'])
                    for key in ['workflow_id', 'workflow_run_id', 'job_id', 'entity_type', 'job_type', 'job_start_time',
                                'job_end_time', 'job_status',
                                'source_name', 'source_file_names', 'source_schema_name', 'source_database_name',
                                'table_group_name',
                                'iwx_table_name', 'starting_watermark_value', 'ending_watermark_value',
                                'target_schema_name', 'target_table_name', 'pre_target_count', 'fetch_records_count',
                                'target_records_count']:
                        od[key] = table.get(key, "")
                    temp.append(od)
            self.job_metrics_final.extend(temp)

    def get_abc_job_metrics(self, time_range_for_jobs_in_mins=5, workflow_id=None, workflow_run_id=None):
        """
        Gets the Infoworks Job metrics
        :param time_range_for_jobs_in_mins: time range to look out for Infoworks jobs (default 5mins)
        :type time_range_for_jobs_in_mins: Integer
        :param workflow_id: Workflow id to get the jobs
        :type workflow_id: String
        :param workflow_run_id: Workflow run id to get the jobs
        :type workflow_run_id: String
        :return: response dict
        """
        try:
            sources_info = self.get_source_info()
            delay = int(time_range_for_jobs_in_mins)
            now = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(minutes=delay)
            date_string = now.strftime('%Y-%m-%dT%H:%M:%SZ')
            source_jobs = self.get_all_source_jobs(date_string)
            source_jobs_source_ids = list(set([i['entity_id'] for i in source_jobs]))
            sources_info = [i for i in sources_info if i['id'] in source_jobs_source_ids]
            with ThreadPoolExecutor(max_workers=10) as executor:
                executor.map(self.get_source_jobs_metrics_results,
                             [date_string] * len(sources_info), sources_info,
                             [workflow_id, workflow_run_id] * len(sources_info))
                executor.shutdown(wait=True)
            pipeline_jobs_list = self.get_pipeline_jobs(date_string)
            with ThreadPoolExecutor(max_workers=10) as executor:
                executor.map(self.get_pipeline_build_metrics_results, pipeline_jobs_list,
                             [workflow_id, workflow_run_id] * len(pipeline_jobs_list))
                executor.shutdown(wait=True)

            result = []
            header = ['workflow_id', 'workflow_run_id', 'job_id', 'entity_type', 'job_type', 'job_start_time',
                      'job_end_time', 'job_status', 'source_name', 'source_file_names', 'source_schema_name',
                      'source_database_name', 'table_group_name', 'iwx_table_name', 'starting_watermark_value',
                      'ending_watermark_value', 'target_schema_name', 'target_table_name', 'pre_target_count',
                      'fetch_records_count', 'target_records_count', 'job_link']
            ui_port = 443 if self.client_config["port"] == '443' else 3000
            if len(self.job_metrics_final) > 0:
                for item in self.job_metrics_final:
                    dict_temp = {}
                    for i in header:
                        dict_temp[i] = item.get(i)
                        if i == "job_link" and dict_temp.get("job_id", None) != None:
                            dict_temp[
                                i] = f"{self.client_config['protocol']}://{self.client_config['ip']}:{ui_port}/job/logs?jobId={dict_temp.get('job_id', '')}"
                    result.append(copy.deepcopy(dict_temp))
                return result
            else:
                print("Job list empty!!!")
        except Exception as e:
            print("Something went wrong")
            print(str(e))
            traceback.print_exc()
